use std::{
    cell::RefCell,
    collections::{HashMap, HashSet},
};

use self::{
    command::{Command, InternalCommand},
    r#trait::{InternalTrait, Trait},
    r#type::InternalType,
    typedef::TypeDef,
};

pub mod command;
pub mod r#trait;
pub mod r#type;
pub mod typedef;

pub struct GlobalNamespace {
    types: RefCell<Vec<InternalType>>,
    traits: RefCell<Vec<InternalTrait>>,
    type_namespace: RefCell<HashMap<String, TypeNamespaceId>>,
    commands: RefCell<Vec<InternalCommand>>,
    data_namespace: RefCell<HashMap<String, DataNamespaceId>>,
}

enum TypeNamespaceId {
    Types(usize),
    Traits(usize),
}

enum DataNamespaceId {
    Commands(usize),
    Globals(usize),
}

impl GlobalNamespace {
    pub fn init() -> Self {
        Self {
            types: RefCell::new(Vec::new()),
            traits: RefCell::new(Vec::new()),
            type_namespace: RefCell::new(HashMap::new()),
            commands: RefCell::new(Vec::new()),
            data_namespace: RefCell::new(HashMap::new()),
        }
    }
    pub fn register_type(&self, name: &str) -> Result<r#type::Type, NsRegistrationError> {
        if self.type_namespace.borrow().contains_key(name) {
            Err(NsRegistrationError::NameAlreadyExists)
        } else {
            self.types.borrow_mut().push(InternalType {
                traits: RefCell::new(HashSet::new()),
                name: name.to_owned(),
            });
            let id = self.types.borrow().len() - 1;
            let _ = self
                .type_namespace
                .borrow_mut()
                .insert(name.to_owned(), TypeNamespaceId::Types(id));
            Ok(r#type::Type {
                id,
                namespace: self,
            })
        }
    }

    pub fn register_trait(&self, name: &str) -> Result<Trait, NsRegistrationError> {
        if self.type_namespace.borrow().contains_key(name) {
            Err(NsRegistrationError::NameAlreadyExists)
        } else {
            self.traits.borrow_mut().push(InternalTrait {
                types: RefCell::new(HashSet::new()),
                name: name.to_owned(),
            });
            let id = self.traits.borrow().len() - 1;
            let _ = self
                .type_namespace
                .borrow_mut()
                .insert(name.to_owned(), TypeNamespaceId::Traits(id));
            Ok(Trait {
                id,
                namespace: self,
            })
        }
    }

    pub fn register_command(
        &self,
        name: &str,
        input: Option<TypeDef>,
        output: Option<TypeDef>,
    ) -> Result<Command, NsRegistrationError> {
        if self.data_namespace.borrow().contains_key(name) {
            Err(NsRegistrationError::NameAlreadyExists)
        } else {
            self.commands.borrow_mut().push(InternalCommand {
                name: name.to_owned(),
                input: input.map(|def| def.into()),
                output: output.map(|def| def.into()),
            });
            let id = self.traits.borrow().len() - 1;
            let _ = self
                .data_namespace
                .borrow_mut()
                .insert(name.to_owned(), DataNamespaceId::Commands(id));
            Ok(Command {
                id,
                namespace: self,
            })
        }
    }

    pub fn get_type(&self, id: usize) -> Option<r#type::Type> {
        if self.types.borrow().len() > id {
            Some(r#type::Type {
                id,
                namespace: self,
            })
        } else {
            None
        }
    }

    pub fn get_trait(&self, id: usize) -> Option<Trait> {
        if self.traits.borrow().len() > id {
            Some(Trait {
                id,
                namespace: self,
            })
        } else {
            None
        }
    }

    pub fn get_command(&self, id: usize) -> Option<Command> {
        if self.commands.borrow().len() > id {
            Some(Command {
                id,
                namespace: self,
            })
        } else {
            None
        }
    }

    pub fn get_type_by_name(&self, name: &str) -> Option<r#type::Type> {
        if let Some(TypeNamespaceId::Types(id)) = self.type_namespace.borrow().get(name) {
            Some(r#type::Type {
                id: *id,
                namespace: self,
            })
        } else {
            None
        }
    }

    pub fn get_trait_by_name(&self, name: &str) -> Option<Trait> {
        if let Some(TypeNamespaceId::Traits(id)) = self.type_namespace.borrow().get(name) {
            Some(Trait {
                id: *id,
                namespace: self,
            })
        } else {
            None
        }
    }

    pub fn get_command_by_name(&self, name: &str) -> Option<Command> {
        if let Some(DataNamespaceId::Commands(id)) = self.data_namespace.borrow().get(name) {
            Some(Command {
                id: *id,
                namespace: self,
            })
        } else {
            None
        }
    }
}

#[derive(Debug)]
pub enum NsRegistrationError {
    NameAlreadyExists,
}
