use codespan_reporting::diagnostic::{Diagnostic, Label};

/// The enum representing a syntax error, used for error reporting
#[derive(Debug, Clone)]
pub enum SyntaxError {
    /// This variant indicates a token that the Lexer didn't recognize
    InvalidToken(Vec<logos::Span>),
    /// `MissingStreamer` means, that the pipeline starts with a Pipe (`|`), so it has no streamer as input in front of it.
    MissingStreamer(Vec<logos::Span>),
    /// `MissingSink` means, that the pipeline ends with a Pipe (`|`), meaning that the output can't go anywhere
    MissingSink(Vec<logos::Span>),
    /// This indicates a missing filter somewhere in the pipeline, meaning that there's 2 pipes after one another
    MissingFilter(Vec<logos::Span>),
    /// A literal cannot be a sink, TODO
    LiteralAsSink(Vec<logos::Span>),
    /// A literal can't be a filter either, TODO
    LiteralAsFilter(Vec<logos::Span>),
    /// A literal acting as streamer cannot take arguments, TODO
    LiteralWithArgs(Vec<logos::Span>),
}

// TODO: much better and more complex errors, with suggestions for fixes
impl SyntaxError {
    pub fn to_diagnostic(&self, file_id: usize) -> Diagnostic<usize> {
        match self {
            Self::InvalidToken(errs) => Diagnostic::error()
                .with_message("failed to parse invalid tokens")
                .with_labels(
                    errs.iter()
                        .map(|span| {
                            Label::primary(file_id, span.clone()).with_message("invalid token")
                        })
                        .collect(),
                ),
            Self::MissingStreamer(locs) => Diagnostic::error()
                .with_message("pipelines must always start with a provider")
                .with_labels(
                    locs.iter()
                        .map(|span| Label::primary(file_id, span.clone()))
                        .collect(),
                ),
            Self::MissingFilter(locs) => Diagnostic::error()
                .with_message("missing filters in pipeline")
                .with_labels(
                    locs.iter()
                        .map(|span| Label::primary(file_id, span.clone()))
                        .collect(),
                ),
            Self::MissingSink(locs) => Diagnostic::error()
                .with_message("pipelines need to end in a sink")
                .with_labels(
                    locs.iter()
                        .map(|span| Label::primary(file_id, span.clone()))
                        .collect(),
                ),
            Self::LiteralAsSink(locs) => Diagnostic::error()
                .with_message("literals cannot be sinks")
                .with_labels(
                    locs.iter()
                        .map(|span| Label::primary(file_id, span.clone()))
                        .collect(),
                ),
            Self::LiteralAsFilter(locs) => Diagnostic::error()
                .with_message("literals cannot be filters")
                .with_labels(
                    locs.iter()
                        .map(|span| Label::primary(file_id, span.clone()))
                        .collect(),
                ),
            Self::LiteralWithArgs(locs) => Diagnostic::error()
                .with_message("literals cannot take arguments")
                .with_labels(
                    locs.iter()
                        .map(|span| Label::primary(file_id, span.clone()))
                        .collect(),
                ),
        }
    }
}
